from __future__ import annotations

import os
from abc import abstractmethod
from datetime import timedelta
from enum import Enum
from typing import Dict, Optional

from dateutil.relativedelta import relativedelta
from redis import Redis

ENABLE_RATE_LIMIT = os.getenv("ENABLE_RATE_LIMIT", "True") == "True"


class RateLimitException(Exception):
    rate_id: str

    def __init__(
        self,
        rate_id: str,
        rate_limit: Optional[float],
        rate_usage: Optional[int],
        hint: str = "",
    ):
        if hint:
            hint = " Hint :: " + hint
        super().__init__(
            f"Usage rate :: {rate_usage} exceeds rate_limit :: {rate_limit} with rate_id :: {rate_id}.{hint}"  # noqa
        )
        self.rate_id = rate_id


class Tier(Enum):
    FREE = "free"
    PRO = "pro"
    MANAGED = "managed"


UNLIMITED = float("inf")


class Counter:
    _redis_client: Redis

    def __init__(
        self,
        timedelta_: timedelta,
        redis_host: str,
        redis_port: int = 6379,
        redis_password: Optional[str] = None,
    ):
        self.redis_host = redis_host
        self.redis_port = redis_port
        self.redis_password = redis_password
        self.timedelta_ = timedelta_

    @property
    def redis_client(self) -> Redis:
        if not hasattr(self, "_redis_client"):
            self._redis_client = Redis(
                host=self.redis_host,
                port=self.redis_port,
                password=self.redis_password,
            )
        return self._redis_client

    def incr(self, key: str) -> None:
        if self.redis_client.exists(key):
            self.redis_client.incr(key)
        else:
            self.redis_client.set(key, 1, ex=self.timedelta_)

    def get(self, key: str) -> int:
        return int(self.redis_client.get(key) or 0)


class ValidateRateLimitRedis:
    def __init__(
        self,
        rate_id: str,
        tier_limits: Dict[Tier, float],
        timedelta_: timedelta,
        redis_host: str,
        redis_port: int = 6379,
        redis_password: Optional[str] = None,
    ):
        assert (
            Tier.FREE in tier_limits and Tier.PRO in tier_limits
        ), f"ValidateRateLimit.tier_limits must declare rate limits for :: {Tier.FREE} and {Tier.PRO}"  # noqa

        self.rate_id = rate_id
        self.counter = Counter(
            timedelta_, redis_host, redis_port, redis_password
        )

        # Set special tier limits
        self.tier_limits = tier_limits
        self.tier_limits[Tier.MANAGED] = UNLIMITED

    def validate_user_rate(self, user_id: str) -> None:
        if not ENABLE_RATE_LIMIT:
            return

        key = f"{user_id}/{self.rate_id}"

        # Get user data
        user_tier = self.get_user_tier(user_id)
        user_rate = self.counter.get(key)

        # Check rate limit
        rate_limit = self.tier_limits.get(user_tier, -1)
        if user_rate >= rate_limit:
            raise RateLimitException(self.rate_id, rate_limit, user_rate)

        # Update user rate
        self.counter.incr(key)

    @abstractmethod
    def get_user_tier(self, user_id: str) -> Tier:
        ...


class ValidateRateLimitPostgres:
    def __init__(
        self,
        rate_id: str,
        tier_limits: Dict[Tier, float],
        window: relativedelta,
    ):
        assert (
            Tier.FREE in tier_limits and Tier.PRO in tier_limits
        ), f"ValidateRateLimit.tier_limits must declare rate limits for :: {Tier.FREE} and {Tier.PRO}"  # noqa

        self.rate_id = rate_id

        # Set special tier limits
        self.tier_limits = tier_limits
        self.tier_limits[Tier.MANAGED] = UNLIMITED
        self.window = window

    def validate_user_rate(self, *, session, user_id: str) -> ValidateRateLimitPostgres:
        if not ENABLE_RATE_LIMIT:
            return

        user_tier = self.get_user_tier(session=session, user_id=user_id)
        user_usage = self.get_user_rate_limit_usage(session=session, user_id=user_id)

        # Check rate limit
        rate_limit = self.tier_limits.get(user_tier, -1)
        if user_usage >= rate_limit:
            raise RateLimitException(self.rate_id, rate_limit, user_usage)

        return self

    @abstractmethod
    def get_user_tier(self, *, session, user_id: str) -> Tier:
        ...

    @abstractmethod
    def get_user_rate_limit_usage(self, *, session, user_id: str) -> int:
        ...

    @abstractmethod
    def increment_user_rate_usage(self, *, session, user_id: str, usage: int) -> int:
        ...


import json
import logging
import os
from contextvars import ContextVar
from typing import Any, Dict, Optional
from uuid import uuid4

import google.cloud.logging
from google.cloud.logging.handlers import CloudLoggingHandler

# Keeps track of variables in the Thread/Task, across different modules.
_binded_vars: ContextVar = ContextVar("Binded vars in this context")


class LoggingManager:
    def __init__(
        self,
        name: str = __name__,
        logger_id: Optional[str] = None,
        gcp_env_variable: str = "GCP_SERVICE_KEY",
    ):
        """
        Args:
            name (str): Name of the logger. Defaults to __name__.
            logger_id (str, optional): id of the logger.
            gcp_env_variable (str): Name of the Google Service Key environment
                variable. The envionrment variable can hold a file path
                or a json.
        """
        if logger_id:
            self._set_logger_id(logger_id)

        self._gcp_client = None
        try:
            if gcp_env_variable in os.environ:
                # Local testing
                self._gcp_client = (
                    google.cloud.logging.Client.from_service_account_json(
                        os.environ[gcp_env_variable]
                    )
                )
            else:
                self._gcp_client = google.cloud.logging.Client()
        except Exception:
            # What this means is that  there is no GCP logging.
            # The most likely reason for this is local development
            # where there is not a Google Service Key.
            pass

        self._logger = logging.getLogger(name)
        self._logger.handlers.clear()
        self._logger.setLevel(logging.DEBUG)

        if self._gcp_client:
            cloudlogging_formatter = logging.Formatter("%(name)s: %(message)s")
            cloud_handler = CloudLoggingHandler(self._gcp_client)
            cloud_handler.setFormatter(cloudlogging_formatter)
            self._logger.addHandler(cloud_handler)

        stream_handler = logging.StreamHandler()
        streamlog_format = "%(_color)s %(asctime)s [%(levelname)s] - %(name)s: %(message)s - JSON Payload: %(json_fields)s \033[0;0m"  # noqa
        streamlog_formatter = logging.Formatter(fmt=streamlog_format)
        stream_handler.setFormatter(streamlog_formatter)
        self._logger.addHandler(stream_handler)

    @property
    def logger_id(self) -> str:
        if "request_id" not in self.binded:
            self._set_logger_id(uuid4().hex)
        return self.binded.get("request_id", "None")

    def _set_logger_id(self, logger_id: str) -> None:
        self.bind(request_id=logger_id)

    @property
    def binded(self) -> Dict[str, Any]:
        try:
            return _binded_vars.get()
        except LookupError:
            _binded_vars.set({})
            return _binded_vars.get()

    def bind(
        self, _payload: Optional[Dict[str, Any]] = None, **kwargs
    ) -> None:
        "Payload is a hack to bind to threads."
        if _payload and not isinstance(_payload, Dict):
            # TODO(luis): Revisit this logic
            self.warning("_payload must be of type Dict. Skipping the bind")
            _payload = {}
        _binded_vars.set((_payload or {}) | kwargs | self.binded)

    def _preprocess_json_payload(
        self,
        msg: str,
        payload: Optional[Dict[str, Any]],
        kwargs: Optional[Dict[str, Any]],
    ) -> Optional[dict]:
        if not payload:
            payload = {}
        if kwargs:
            payload = payload | kwargs
        payload = payload | self.binded

        payload["_message"] = msg
        payload["request_id"] = self.logger_id
        return payload

    def log(
        self,
        msg: str,
        level: int,
        json_params: Optional[Dict[str, Any]],
        skip_if_local: bool,
        color: str = "\033[0;0m",
        **kwargs,
    ) -> None:
        if skip_if_local and not self._gcp_client:
            return

        json_params = self._preprocess_json_payload(msg, json_params, kwargs)
        if self._gcp_client:
            try:
                self._logger.log(
                    level,
                    msg,
                    extra={"json_fields": json_params, "_color": color},
                )
            except Exception as e:
                self._logger.warning(
                    f"Error serializing JSON log: {e}",
                    extra={
                        "json_fields": {"request_id": self.logger_id},
                        "_color": "\033[0;33m",
                    },
                )
        else:
            try:
                self._logger.log(
                    level,
                    msg,
                    extra={
                        "json_fields": json.dumps(json_params, indent=2),
                        "_color": color,
                    },
                )
            except TypeError as e:
                # TODO(luis): Find a better way to pass color.
                self._logger.warning(
                    f"Error serializing JSON log: {e}",
                    extra={
                        "json_fields": {"request_id": self.logger_id},
                        "_color": "\033[0;33m",
                    },
                )
                self._logger.log(
                    level,
                    msg,
                    extra={"json_fields": json_params, "_color": color},
                )

    def debug(
        self,
        msg: str,
        json_params: Optional[Dict[str, Any]] = None,
        skip_if_local: bool = False,
        **kwargs,
    ) -> None:
        self.log(
            msg,
            level=logging.DEBUG,
            json_params=json_params,
            skip_if_local=skip_if_local,
            color="\033[0;32m",
            **kwargs,
        )

    def info(
        self,
        msg: str,
        json_params: Optional[Dict[str, Any]] = None,
        skip_if_local: bool = False,
        **kwargs,
    ) -> None:
        self.log(
            msg,
            level=logging.INFO,
            json_params=json_params,
            skip_if_local=skip_if_local,
            color="\033[0;0m",
            **kwargs,
        )

    def warning(
        self,
        msg: str,
        json_params: Optional[Dict[str, Any]] = None,
        skip_if_local: bool = False,
        **kwargs,
    ) -> None:
        self.log(
            msg,
            level=logging.WARNING,
            json_params=json_params,
            skip_if_local=skip_if_local,
            color="\033[0;33m",
            **kwargs,
        )

    def error(
        self,
        msg: str,
        json_params: Optional[Dict[str, Any]] = None,
        skip_if_local: bool = False,
        **kwargs,
    ) -> None:
        self.log(
            msg,
            level=logging.ERROR,
            json_params=json_params,
            skip_if_local=skip_if_local,
            color="\033[0;31m",
            **kwargs,
        )

    @property
    def gcp_logging_client(self) -> Optional[google.cloud.logging.Client]:
        "WARNING: To be deprecated."
        return self._gcp_client
